/*
 *	Source file	: mp3_14_menu.c
 *	Date		: 2019/06/15
 *	Update		: 2019/12/14
 *	Update		: 2020/05/03 (インターネットラジオ IR01 対応)
 */

#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <wiringPi.h>
#include <wiringSerial.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/time.h>
#include <signal.h>
#include <time.h>
#include <sys/socket.h>
#include <poll.h>
#include <sys/stat.h>

#include "mp3_14.h"

// 入出力のピン番号をwiringPIのピン番号で定義
// 1		GPIO18
// 4		GPIO23
// 5		GPIO24
// 6		GPIO25
// 10		GPIO8
// 11		GPIO7
// 26		GPIO12
// 27		GPIO16
// 9		GPIO3
// 8		GPIO2
// 29		GPIO21
#define PIN_OUT_ACTIVE			16		// GPIO15 ZERO実行状態
#define PIN_OUT_MUTE			15		// GPIO14 MUTE
#define PIN_INP_PDOWN			10		// GPIO8 PICからのPOWER DOWN指示

#define	ZERO_IS_ACTIVE			1
#define	ZERO_IS_INACTIVE		0

#define MUTE_OUTPUT_ON			1
#define MUTE_OUTPUT_OFF			0

// 機能
#define PLAY_NEXT_MP3           0x91
#define PLAY_PREV_MP3           0x92
#define PLAY_OTHER_DIR          0x93
#define POWEROFF                0x94
#define GOTO_MAINMENU           0x95
#define SELECT_OTHER_MP3		0x96
#define	PLAY_ONCEMORE			0x97
#define HOST_HALT				0x98

// 戻り値
#define EXIT_SHUTDOWN			0x00
#define	EXIT_HALT				0x01

// ウインドウズ版コントローラとの通信コマンドコード (PinEventの取りえる値)
#define	WINCTRL_CMD_PLAYDIR		0x01
#define	WINCTRL_CMD_PLAY		0x02
#define	WINCTRL_CMD_PLAYRANDOM	0x03
#define	WINCTRL_CMD_VOLUP		0x04
#define	WINCTRL_CMD_VOLDOWN		0x05
#define	WINCTRL_CMD_PLAYSTOP	0x06
#define	WINCTRL_CMD_PLAYSKIP	0x07
#define	WINCTRL_CMD_ACTIVE		0x08
#define	WINCTRL_CMD_NOWPLAYING	0x09
#define	WINCTRL_CMD_WAITING		0x0A
#define	WINCTRL_CMD_NEXT		0x0B
#define	WINCTRL_CMD_PREV		0x0C
#define	WINCTRL_CMD_SHUTDOWN	0x0D
#define WINCTRL_CMD_FILESIZE	0x0E
#define WINCTRL_CMD_EXECMKFL	0x0F
#define WINCTRL_CMD_DATATIMESTAMP	0x10
#define WINCTRL_CMD_HALT		0x11
#define IR01_CMD_PLAYONE		0x20
#define	IR01_CMD_EXISTMP3		0x21

// ボタン操作/イベント (PinEventの取りえる値：WINCTRL_CMD_***と値が重複しないこと)
#define PDOWN_PRESSED			0x80

#define	NOWPLAYING_STARTEDNOW	0x01
#define	NOWPLAYING_CONTINUE		0x00

#define MIN_VOLUME				-175	                // MP3デコーダ madplay の制限
#define MAX_VOLUME				+18     	            // MP3デコーダ madplay の制限

// madplayへ送信するコマンド(すべてをサポートしてない)
// madplay-0.15.2bのplayer.cから流用
enum {
  KEY_PAUSE    = 'p',
  KEY_STOP     = 's',
  KEY_FORWARD  = 'f',
  KEY_BACK     = 'b',
  KEY_TIME     = 't',
  KEY_QUIT     = 'q',
  KEY_INFO     = 'i',
  KEY_GAINDECR = '-',
  KEY_GAININCR = '+',
  KEY_GAINZERO = '_',
  KEY_GAININFO = '='
 ,KEY_GETPOS   = 'z'			// MP3_14で追加
};

// その他定数
#define	AUTOPOFF_PLAYING_TIMEOUT		(4*60*60*1000)		// 再生中の自動シャットダウン 4hours
#define	AUTOPOFF_SELECT_TIMEOUT			(3*60*1000)			// 選択操作中の自動シャットダウン時間 3min
#define	LONGPRESS_INTERVAL				(1500)				// ボタン長押し時間 1.5sec
#define	PLAYTIME_INTERVAL				(5*1000)			// 演奏時間表示更新時間間隔 5sec
#define NOTIFY_ACTIVE_HOST_INTERVAL		(5*1000)			// HOST起動中をコントローラに通知する時間間隔 5sec
#define	DEFAULT_VOLUME					0
#define	VOLUME_STEP						4					// 音量の増減単位
#define	WINCTRL_CMDLENGTH				16					// コントローラとの通信１単位のバイト数
#define PLAYER_CMDLENGTH				5					// madplayとの通信１単位のバイト数
#define CMD_BUF_SIZE					16					// WINCTRL_CMDLENGTH,PLAYER_CMDLENGTH以上のサイズ
#define POLL_TIMEOUT					50					// poll関数のタイムアウト (ms)

// 外部プログラム
#define COMMUNICATION_SERVER	"./mp3_14_serv"
#define FIND_SND_ROOT			"find %s -maxdepth 10 -type d > %s"
#define MAKE_SNDFILE_LIST		"./mp3_14_mkfl %s"

// イベント/制御
static volatile int	PinEvent;
static unsigned int AutoPoffTimer;
unsigned int 		LastEventTime;
int					fdWin[2];			// mp3_14_servとの通信(socketpair)
pid_t				PidWinServ;			// mp3_14_servのプロセスID

// 再生モード：選択して再生か、ランダム再生か
typedef enum {PM_RANDOM_DIR, PM_SELECT} playmode_t;
static playmode_t   PlayMode;

// 再生単位：ディレクトリ/ファイル
typedef enum {PU_DIR, PU_FILE} playunit_t;

// PlayMp3InSubdirectoryの動作状態
typedef enum {PS_PLAY, PS_SELECT} playorselect_t;

// 音量
static int		Volume;
// データファイルのタイムスタンプ
static char			DataTimeStamp[DL_CRDATETIME_LEN + 1];

// ワーク
static uint8_t      PlayDirHist[RECLEN_HISTFILE];			// 履歴ファイルレコード

// 関数宣言
void InitApp(uint8_t *RecDirList);
void ExecWinServProc();
void TestWinServProc();
void PIN_IRQHandler(void);
int PlayRandom(uint8_t *RecDirList, uint32_t restart_time);
int ReadWindowsController(unsigned char *buf, int buf_size);
int WindowsController(uint8_t *RecDirList);
int PlayOneDirectory(uint8_t *RecDirList, uint16_t dirno, uint8_t start_trackno, uint32_t start_time);
int PlayMp3InSubdirectory(uint8_t *RecDirList, uint16_t dirno, uint16_t numsndfile, 
		uint8_t start_trackno, uint32_t start_time, playorselect_t playorselect, playunit_t playunit);
int PlayOneMp3(const char *mp3file, uint16_t dirno, uint8_t trackno, uint32_t start_time);
int PlayOneMp3_IR01(const char *mp3file, uint32_t start_time);
void SendCmdNoMp3_IR01(char exist_flg);
void SendCmdToCtrlNowPlaying(uint16_t dirno, uint8_t trackno, unsigned int current_play_time, int volume, uint8_t flag, unsigned int position);
void SendCmdToCtrlFilesize(unsigned int filesize);
void ReadHistory();
void SaveHistory();
int writeall(int fd, void *buf, size_t nbyte);
int readall(int fd, void *buf, size_t nbyte);
void ExecMkfl(unsigned char *param);
void ReadDataTimestamp(uint8_t *RecDirList);
void SendDataTimestamp();


int main(){
    uint8_t     RecDirList[RECLEN_DIRLIST_FULL];
    int			ret = EXIT_HALT;
    
	InitApp(RecDirList);
	ReadHistory();
	
	ret = WindowsController(RecDirList);
//	TestWinServProc();
	
	return ret;
}


/*
 *	初期化処理
 *	1.WiringpiによるIO設定
 *	2.mp3_14_serv起動
 *	3.データファイルのタイムスタンプ取得
 */
void InitApp(uint8_t *RecDirList){
	LastEventTime = 0;
	wiringPiSetup();
	
	pinMode(PIN_INP_PDOWN, INPUT);
	pullUpDnControl(PIN_INP_PDOWN, PUD_UP);
	//waitForInterrupt(PIN_INP_PDOWN, -1);
	//wiringPiISR(PIN_INP_PDOWN, INT_EDGE_FALLING, PIN_IRQHandler);
	
	pinMode(PIN_OUT_ACTIVE, OUTPUT);
	digitalWrite(PIN_OUT_ACTIVE, ZERO_IS_ACTIVE);
	
	pinMode(PIN_OUT_MUTE, OUTPUT);
	digitalWrite(PIN_OUT_MUTE, MUTE_OUTPUT_OFF);
	
	ExecWinServProc();
	
	ReadDataTimestamp(RecDirList);
}


/*
 *	Windows版コントローラとの通信プロセス(mp3_14_serv)を起動
 */
void ExecWinServProc(){
	char	*args[16];			// madplay起動時引数配列
	int		argidx;
	char	fdstr[24];
	
	if (socketpair(AF_LOCAL, SOCK_STREAM, 0, fdWin) == -1){
		printf("socketpair error %d\n", errno);
		return;
	}
	printf("fdWin=%d, %d\n", fdWin[0], fdWin[1]);
	
	PidWinServ = fork();
	switch(PidWinServ){
	case -1:
		exit(1);
		break;
	case 0:
		argidx = 0;
		args[argidx++] = "mp3_14_serv";
		sprintf(fdstr, "%d", fdWin[1]);
		args[argidx++] = fdstr;
		args[argidx++] = (char *)0;
		execv(COMMUNICATION_SERVER, args);
		_exit(errno);
		break;
	default:
		break;
	}
}


void TestWinServProc(){
	char	buf[1024];
	int		rval;
	
	while (1){
		if ((rval = read(fdWin[0], buf, WINCTRL_CMDLENGTH)) > 0){
			write(fdWin[0], buf, rval);
		}
	}
}


// Interrupt handler
// プログラム起動時、予想外にイベントが発生していることがある？ ので、この関数は使わない。(2019/12/08)
void PIN_IRQHandler(void){
	unsigned int e = millis();
	if ((e - LastEventTime) > 600){			// チャッタリング対策
		if (digitalRead(PIN_INP_PDOWN) == 0){
			printf("digitalRead=PIN_INP_PDOWN\n");
			PinEvent = PDOWN_PRESSED;
		}
		
		if (PinEvent != 0){
			printf("event=%d\n", PinEvent);
		}
		LastEventTime = e;
	}
}


/*
 *	Windows版コントローラ
 */
int WindowsController(uint8_t *RecDirList){
	unsigned char	buf[WINCTRL_CMDLENGTH];
	uint16_t dirno;
	uint8_t trackno;
	uint32_t start_time;
	int	ret = HOST_HALT;
	int loop_continue = 1;
	unsigned int	timeout;
	unsigned int	current_time;
	unsigned int	notify_time;
	
	printf("WindowsController start\n");
	current_time = millis();
	timeout = current_time + AUTOPOFF_SELECT_TIMEOUT;
	notify_time = current_time + NOTIFY_ACTIVE_HOST_INTERVAL;
	while (loop_continue){
		if (ReadWindowsController(buf, WINCTRL_CMDLENGTH) > 0){
			printf("WindowsController cmd=%x\n", buf[0]);
			switch (buf[0]){
			case WINCTRL_CMD_PLAYDIR:			// コントローラで選択したディレクトリを再生開始要求
				AutoPoffTimer = millis() + AUTOPOFF_PLAYING_TIMEOUT;
				dirno = (buf[1] | ((uint16_t)buf[2] << 8));
				trackno = buf[3];
				start_time = (buf[4] | ((uint16_t)buf[5] << 8));
				ret = PlayOneDirectory(RecDirList, dirno, trackno, start_time);
	            if (ret == POWEROFF){
					loop_continue = 0;
	            }
				break;
			case WINCTRL_CMD_PLAYRANDOM:		// ランダム再生要求
				start_time = (buf[1] | ((uint16_t)buf[2] << 8));
				ret = PlayRandom(RecDirList, start_time);
	            if (ret == POWEROFF){
					loop_continue = 0;
	            }
				break;
			case WINCTRL_CMD_SHUTDOWN:			// 停止要求
				ret = POWEROFF;
				loop_continue = 0;
				break;
			case WINCTRL_CMD_HALT:				// 終了要求 (電源オフなし)
				ret = HOST_HALT;
				loop_continue = 0;
				break;
			case WINCTRL_CMD_EXECMKFL:			// データファイル作成要求
				ExecMkfl(buf + 1);
				ReadDataTimestamp(RecDirList);
				memset(&PlayDirHist, 0, RECLEN_HISTFILE);
				break;
			case WINCTRL_CMD_DATATIMESTAMP:		// データファイルのタイムスタンプ要求
				SendDataTimestamp();
				break;
			case IR01_CMD_PLAYONE:				// IR01からのMP3ファイル再生要求 (2020/05/03)
				{
					char	mp3filepath[64];
					char	name[9];
					
					memcpy(name, buf+1, 8);
					name[8] = '\0';
					sprintf(mp3filepath, "%s/%s.mp3", SND_IR01, name);
					start_time = (buf[9] | ((uint16_t)buf[10] << 8));
					ret = PlayOneMp3_IR01(mp3filepath, start_time);
		            if (ret == POWEROFF){
						loop_continue = 0;
		            }
				}
				break;
			}
			// ウインドウズ版コントローラからアクセスがあればタイムアウトを更新
			timeout = millis() + AUTOPOFF_SELECT_TIMEOUT;
		}
		
		// PICからの終了指示
#if 0	// 未完 (2019/12/08)
		if (digitalRead(PIN_INP_PDOWN) == 0){
			putchar('0');
			ret = POWEROFF;
			loop_continue = 0;
		} else {
		//	putchar('1');
		}
#endif
		
		current_time = millis();
		if (current_time > timeout){		// ウインドウズ版コントローラからのアクセスが所定時間以上ない
			ret = POWEROFF;
			break;
		} else if (current_time > notify_time){
//			printf("WINCTRL_CMD_WAITING\n");
			buf[0] = WINCTRL_CMD_WAITING;
			writeall(fdWin[0], buf, WINCTRL_CMDLENGTH);
			notify_time = current_time + NOTIFY_ACTIVE_HOST_INTERVAL;
		}
	}
	
	printf("WindowsController stop\n");
	digitalWrite(PIN_OUT_MUTE, MUTE_OUTPUT_OFF);
	
	// Windows版コントローラに停止通知
	buf[0] = WINCTRL_CMD_SHUTDOWN;
	writeall(fdWin[0], buf, WINCTRL_CMDLENGTH);
	sleep(1);
	
	// mp3_14_servを停止
	kill(PidWinServ, SIGKILL);
	
	// プログラムの終了コード
    if (ret == POWEROFF){
		// PICに通知
		digitalWrite(PIN_OUT_ACTIVE, ZERO_IS_INACTIVE);
		printf("mp3_14_menu pdown\n");
    	return EXIT_SHUTDOWN;
    } else {
		printf("mp3_14_menu halt\n");
    	return EXIT_HALT;
    }
}


/*
 *	Windows版コントローラからディレクトリを選択して再生
 */
int PlayOneDirectory(uint8_t *RecDirList, uint16_t dirno, uint8_t start_trackno, uint32_t start_time){
	FILE		*Fil;
    uint32_t    ofs;
    uint16_t    numsnds;
    playunit_t	playunit;
    int			ret = 0;
	
	Fil = fopen(DIRSTRUCTUREFILE14, "r");
	if (Fil == NULL){
		return 0;
	}
    ofs = (uint32_t)dirno * RECLEN_DIRLIST_FULL;
	if (fseek(Fil, ofs, SEEK_SET) == 0){
		if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) == RECLEN_DIRLIST_FULL){
	    	numsnds = *(RecDirList + DL_NUM_SNDFILE_OFFSET);    // number of sound files
		    playunit = PU_DIR;
            ret = PlayMp3InSubdirectory(RecDirList, dirno, numsnds, start_trackno, start_time, PS_PLAY, playunit);
            if (ret == POWEROFF){
            }
	    	
	    }
	}
    fclose(Fil);
	return ret;
}


/*
 *	ランダム再生
 *		restart_time:最初の曲の再生開始秒数（再生の一時停止→再開、で使う）
 */
int PlayRandom(uint8_t *RecDirList, uint32_t restart_time){
	FILE		*Fil;
	FILE		*FilRnd;
    int			ret;
    uint16_t    numsnds;
    uint32_t    ofs;
	uint16_t	rndrecno;
	uint8_t		rndtype;
    uint16_t    dirno;
    uint8_t		start_trackno;
	uint32_t	start_time;
    uint8_t		recrndorder[RECLEN_RANDOMORDER];
    playunit_t	playunit;
    int			first_flg = 1;
    int			restart_seldir = 0;
	
	ret = GOTO_MAINMENU;
	
	FilRnd = fopen(RANDOMORDERFILE, "r");
	if (FilRnd == NULL){
		return 0;
	}
	Fil = fopen(DIRSTRUCTUREFILE14, "r");
	if (Fil == NULL){
		return 0;
	}
	
	memcpy(&dirno, PlayDirHist + HS_LASTDIR_OFFSET, HS_LASTDIR_LEN);
	memcpy(&rndrecno, PlayDirHist + HS_RANDRECNO_OFFSET, HS_RANDRECNO_LEN);
    start_trackno = PlayDirHist[HS_LASTTRACK_OFFSET];
    if (restart_time > 0){
    	start_time = restart_time;
    } else {
		memcpy(&start_time, PlayDirHist + HS_LASTTIME_OFFSET, HS_LASTTIME_LEN);
    }
	AutoPoffTimer = millis() + AUTOPOFF_PLAYING_TIMEOUT;
	printf("dirno=%x rndrecno=%x\n", dirno, rndrecno);
	
	while (1){
		if (first_flg == 1){	// ランダム再生開始
			first_flg = 0;
			// 履歴に記録された中断箇所からの再開
			
		    if (fseek(FilRnd, rndrecno*RECLEN_RANDOMORDER, SEEK_SET) == 0
		    		&& fread(recrndorder, 1, RECLEN_RANDOMORDER, FilRnd) == RECLEN_RANDOMORDER){
		    	uint16_t	rnddirno;
		    	memcpy(&rnddirno, recrndorder + RO_DIRNO_OFFSET, RO_DIRNO_LEN);
		    	printf("  dirno=%x rnddirno=%x\n", dirno, rnddirno);
		    	if (rnddirno != dirno && dirno != 0){
		    		// 履歴ファイルの「再生中のディレクトリの番号」は「ランダム再生順レコード番号」の指示するディレクトリ番号と異なる。
		    		// この場合、フォルダ選択からの再生の途中と考えられる。
		    		// ただし dirno==0のときは履歴データの初期値であってフォルダ選択の再生ではない。
		    		// ->ランダム再生順ファイルから最初の再生対象を取得（次のelseブロック）
				    playunit = PU_DIR;
				    restart_seldir = 1;
				} else {
					// ランダム再生が中断中。
				    rndtype = recrndorder[0];
				    dirno = rnddirno;		// dirno==0のときの対応
				    if (rndtype == RANDOMTYPE_FOLDER){
					    playunit = PU_DIR;
					} else {
					    playunit = PU_FILE;
					}
		    	}
		    } else {
		    	continue;
		    }
		} else {
		    if (fseek(FilRnd, rndrecno*RECLEN_RANDOMORDER, SEEK_SET) == 0
		    		&& fread(recrndorder, 1, RECLEN_RANDOMORDER, FilRnd) == RECLEN_RANDOMORDER){
			    rndtype = recrndorder[0];
			    if (rndtype == RANDOMTYPE_FOLDER){
				    playunit = PU_DIR;
				} else {
				    playunit = PU_FILE;
				}
			    memcpy(&dirno, recrndorder + RO_DIRNO_OFFSET, RO_DIRNO_LEN);
				if (rndtype == RANDOMTYPE_FILE){
			   		start_trackno = recrndorder[RO_TRACKNO_OFFSET];
			   		start_time = 0;
			    } else {
			    	start_trackno = 0;
			   		start_time = 0;
			    }
		    } else {
		    	if (rndrecno == 0) break;
		    	rndrecno = 0;
		    	continue;
		    }
		}
		
		printf("first_flg=%d restart_seldir=%d dirno=%uld \n", first_flg, restart_seldir, dirno);
	    // ディレクトリ構成ファイル
        ofs = (uint32_t)(dirno) * (uint32_t)RECLEN_DIRLIST_FULL;
        if (fseek(Fil, ofs, SEEK_SET) == 0){
        	if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) == RECLEN_DIRLIST_FULL){
        		//printf("rnd = dirno=%x (%d)(ofs=%lx) trackno=%d [%s]\n", dirno, dirno, dirno*RECLEN_DIRLIST, start_trackno, RecDirList + DL_FULLDIR_OFFSET);
                numsnds = *(RecDirList + DL_NUM_SNDFILE_OFFSET);    // number of sound files
                if (numsnds > 0){
                    if (!(start_trackno >= 0 && start_trackno < numsnds)){
                        start_trackno = 0;
                    }
                    // ファイル選択からの再生の再開のときは"ランダム再生順レコード番号","ランダムフラグ"をセットしない。
                    if (restart_seldir == 0){
	                    memcpy(PlayDirHist + HS_RANDRECNO_OFFSET, &rndrecno, HS_RANDRECNO_LEN);
	                }
                    //
                    ret = PlayMp3InSubdirectory(RecDirList, dirno, numsnds, start_trackno, start_time, PS_PLAY, playunit);
                    if (ret == POWEROFF){
                        break;
                    } else if (ret == GOTO_MAINMENU){
                        break;
                    }
                } else {
                }
        	}
        }
        
		if (restart_seldir == 1){
			restart_seldir = 0;
        } else {
	        rndrecno++;
	    }
	}
	
    fclose(Fil);
    fclose(FilRnd);
    
    return ret;
}


/*
 *	mp3_14_serv(Windows版コントローラ)からの通信を取得。
 *	戻り値：取得バイト数。取得できないとき(受信データが無いとき)は 0。エラー時は -1
 */
int ReadWindowsController(unsigned char *buf, int buf_size){
	struct pollfd fd_poll[1];
	int	rval = 0;
	int	total = 0;
	
	while (total != buf_size){
		memset(&fd_poll, 0, sizeof(fd_poll));
		fd_poll[0].fd = fdWin[0];
		fd_poll[0].events = POLLIN | POLLERR;
		poll(fd_poll, 1, POLL_TIMEOUT);
		if (fd_poll[0].revents & POLLIN) {
			rval = (int)readall(fdWin[0], buf + total, buf_size - total);
			if (rval == -1 || rval == 0) break;
			total += rval;
		} else {
			if (total == 0){
				break;
			}
		}
	}
	return rval;
}


/*
 *	指定ディレクトリ内の曲を順次再生する
 *	start_trackno	再生開始トラック番号
 *	start_time		曲途中から再生再開する場合の再生開始時間。最初の曲にだけ適用。
 *	playorselect	トラックの選択処理から始めるか、指定トラックの再生から始めるか
 *	playunit		PU_DIR:フォルダ内全曲再生、PU_FILE:指定トラック(start_trackno)のみ再生
 */
int PlayMp3InSubdirectory(uint8_t *RecDirList, uint16_t dirno, uint16_t numsndfile, 
		uint8_t start_trackno, uint32_t start_time, playorselect_t playorselect, playunit_t playunit){
    int		    ret;
    FILE		*filML;
    uint8_t     RecMp3List[RECLEN_MP3LIST];   // for MP3 list file
    uint8_t     trackno;		// 再生中の曲番 (0,1,2,...,numsndfile-1)
    uint16_t	recno1st;		// 1曲目のMP3ファイルリストレコード番号(0,1,2,…)
    int		 	volinit = 0;
    uint16_t    dirlen;
    uint8_t     execplayflg = 0;
	
    memcpy(&dirlen, RecDirList + DL_LENFULLDIR_OFFSET, DL_LENFULLDIR_LEN);
    memcpy(&recno1st, RecDirList + DL_RECNO_SNDFILE_OFFSET, DL_RECNO_SNDFILE_LEN);
    
    // Volume file
    *(RecDirList + DL_FULLDIR_OFFSET + dirlen) = '/';
    strcpy((char *)(RecDirList + DL_FULLDIR_OFFSET + dirlen + 1), VOLUMEFILE);
    filML = fopen((char *)(RecDirList + DL_FULLDIR_OFFSET), "r");
    if (filML != NULL){
    	if (fread(&Volume, sizeof(int), 1, filML) != 1){
		    Volume = DEFAULT_VOLUME;
    	}
    	fclose(filML);
    } else {
	    Volume = DEFAULT_VOLUME;
    }
	volinit = Volume;
	printf("Volume=%d\n", Volume);
	
    // MP3 file list
    *(RecDirList + DL_FULLDIR_OFFSET + dirlen) = '/';
//    strcpy((char *)(RecDirList + DL_FULLDIR_OFFSET + dirlen + 1), SNDFILELISTFILE);
//    filML = fopen((char *)(RecDirList + DL_FULLDIR_OFFSET), "r");
    filML = fopen(SNDFILELISTFILE, "r");	// カレントディレクトリに存在する
    if (filML == NULL) return 0;
    
    if (start_trackno >= numsndfile) start_trackno = 0;
    trackno = start_trackno;
	digitalWrite(PIN_OUT_MUTE, MUTE_OUTPUT_ON);
    
    while (1){
    	printf(" trackno=%d numsndfile=%d \n", trackno, numsndfile);
    	if (fseek(filML, (uint32_t)(recno1st + trackno)*(RECLEN_MP3LIST), SEEK_SET) != 0) break;
    	if (fread(RecMp3List, 1, RECLEN_MP3LIST, filML) != RECLEN_MP3LIST) break;
        
        // Mp3 file full path
        *(RecDirList + DL_FULLDIR_OFFSET + dirlen) = '/';
        strcpy((char *)(RecDirList + DL_FULLDIR_OFFSET + dirlen + 1), (char *)(RecMp3List + ML_FILENAME_OFFSET));
        // Play a MP3 file
        ret = PlayOneMp3((char *)(RecDirList + DL_FULLDIR_OFFSET), dirno, trackno, start_time);
        if (ret == PLAY_OTHER_DIR){
            break;
        } else if (ret == POWEROFF){
            break;
        } else if (ret == GOTO_MAINMENU){
            break;
        } else if (ret == SELECT_OTHER_MP3){
            if (PlayMode == PM_RANDOM_DIR){
            	ret = GOTO_MAINMENU;
            	break;
            } else {
                playorselect = PS_SELECT;
            }
        } else if (ret == PLAY_PREV_MP3){
            if (trackno > 0){
            	trackno--;
            	start_time = 0;
            }
        } else {
            execplayflg = 1;
            trackno++;
            start_time = 0;
            if (trackno == numsndfile || playunit == PU_FILE) break;
        }
    }
    
    fclose(filML);
	digitalWrite(PIN_OUT_MUTE, MUTE_OUTPUT_OFF);
    
    // VOLUMEFILE保存
    //if (execplayflg == 1 && Volume != volinit){     // すくなくとも１曲再生完了したとき
    {		// mp3_14では1曲目の途中であっても保存する
	    *(RecDirList + DL_FULLDIR_OFFSET + dirlen) = '/';
	    strcpy((char *)(RecDirList + DL_FULLDIR_OFFSET + dirlen + 1), VOLUMEFILE);
	    filML = fopen((char *)(RecDirList + DL_FULLDIR_OFFSET), "w");
	    if (filML != NULL){
	    	fwrite(&Volume, sizeof(int), 1, filML);
	    	fclose(filML);
	    }
    }
    
    return ret;
}


/*
 * MP3再生
 * 	mp3file		MP3ファイルフルパス
 *	dirno		ディレクトリ構成ファイルのレコード番号（ルートディレクトリ=1から始まる行番号）
 *	trackno		0から始まる曲番号
 *	start_time	再生開始時間(秒)
 *	
 */
int PlayOneMp3(const char *mp3file, uint16_t dirno, uint8_t trackno, uint32_t start_time){
	pid_t	pid;
	int 	status;
	int		fdplayer[2];
	char	cmd[CMD_BUF_SIZE];
	char	fdstr[24];
	char	attenuate[24];
	char	start[24];
	//char	quiet[16];
	char	notty[24];
	//char	displaytime[32];
	char	*args[16];			// madplay起動時引数配列
	int		argidx;
    int		ret;
    int		volstep;
	unsigned int	start_systime_ms;		// 再生開始システム時刻(ミリ秒：millis()の戻り値)
	unsigned int	current_systime_ms;		// 現在システム時刻(ミリ秒：millis()の戻り値)
	unsigned int 	total_play_time;		// 曲の長さ（秒）
	unsigned int 	current_play_time;		// 現在の再生秒
	int		send_now_playing;
	struct stat		mp3stat;				// 
	int		cmd_acceptable;					// 1:デコーダへのコマンド可能
	
	printf("PlayOneMp3=[%s] vol=%d start_time=%x \n", mp3file, Volume, (int)start_time);
restart_label:

	if (stat(mp3file, &mp3stat) < 0){
		return PLAY_NEXT_MP3;
	}
	
	// 履歴レコード
    memcpy(PlayDirHist + HS_LASTDIR_OFFSET, &dirno, HS_LASTDIR_LEN);
	PlayDirHist[HS_LASTTRACK_OFFSET] = trackno;
	memcpy(PlayDirHist + HS_LASTTIME_OFFSET, &start_time, HS_LASTTIME_LEN);
	SaveHistory();
	
	if (socketpair(AF_LOCAL, SOCK_STREAM, 0, fdplayer) == -1){
		printf("socketpair error %d\n", errno);
		return PLAY_NEXT_MP3;
	}
	//printf("fd=%d, %d\n", fdplayer[0], fdplayer[1]);
	
	pid = fork();
	switch(pid){
	case -1:
		exit(1);
		break;
	case 0:
		argidx = 0;
		args[argidx++] = "madplay";
		args[argidx++] = (char *)mp3file;
		sprintf(fdstr, "--socket=%d", fdplayer[1]);
		args[argidx++] = fdstr;
		if (Volume < 0){
			if (Volume < MIN_VOLUME) Volume = MIN_VOLUME;
			sprintf(attenuate, "--attenuate=%d", Volume);
			args[argidx++] = attenuate;
		} else if (Volume > 0){
			if (Volume > MAX_VOLUME) Volume = MAX_VOLUME;
			sprintf(attenuate, "--amplify=%d", Volume);
			args[argidx++] = attenuate;
		}
//		if (start_time > 0){
		{
			int h, m, s;
			h = (start_time / 3600);			// hour
			m = (start_time - h*3600) / 60;	// min
			s = (start_time % 60);
			sprintf(start, "--start=%02d:%02d:%02d.000", h, m, s);
			args[argidx++] = start;
		}
//		sprintf(quiet, "--very-quiet");
//		sprintf(quiet, "--verbose");		// madplayのshow_status()を実行させる
//		args[argidx++] = quiet;
		sprintf(notty, "--no-tty-control");
		args[argidx++] = notty;
//		sprintf(displaytime, "--display-time=%s", "remaining");		// remaining, current, overall
//		args[argidx++] = displaytime;
		args[argidx++] = (char *)0;
		//for (i=0; i<(argidx - 1); i++){
		//	printf("arg=[%s]\n", args[i]);
		//}
		//execl("/usr/local/bin/madplay", "madplay", mp3file, fdstr, stereo, attenuate, (char *)0);
		execv("/usr/local/bin/madplay", args);
		_exit(errno);
		break;
	default:
		SendCmdToCtrlNowPlaying(dirno, trackno, start_time, Volume, NOWPLAYING_STARTEDNOW, 0);
		//printf("madplay started pid=%d\n", pid);
		start_systime_ms = millis();
		ret = 0;
		PinEvent = 0;
		send_now_playing = 0;
		cmd_acceptable = 1;		// デコーダにコマンド送信可能
		
		do {
			unsigned char	buf[CMD_BUF_SIZE];
			
			// ウインドウズ版コントローラからの受信
			if (ReadWindowsController(buf, WINCTRL_CMDLENGTH) > 0 && ret == 0){
				// retの値はMP3_14/decoder.c(run_sync())に合わせる
				if (cmd_acceptable == 1){
					PinEvent = buf[0];
					//printf("cmd from client = %x\n", PinEvent);
					if (PinEvent == WINCTRL_CMD_VOLUP ){
						//volstep = MAX_VOLUME - Volume;
						//if (volstep > VOLUME_STEP) volstep = VOLUME_STEP;
						volstep = buf[1];
						if ((Volume + volstep) > MAX_VOLUME) volstep = MAX_VOLUME - Volume;
						if (volstep > 0){
							cmd[0] = KEY_GAININCR;
							cmd[1] = volstep;			// VOLUME_STEP
		    				writeall(fdplayer[0], cmd, 2);
		    				Volume += volstep;			// VOLUME_STEP;
		    				if (send_now_playing == 0) send_now_playing = 1;
		    			}
					} else if (PinEvent == WINCTRL_CMD_VOLDOWN ){
						//volstep = Volume - MIN_VOLUME;
						//if (volstep > VOLUME_STEP) volstep = VOLUME_STEP;
						volstep = buf[1];
						if ((Volume - volstep) < MIN_VOLUME) volstep = Volume - MIN_VOLUME;
						if (volstep > 0){
							cmd[0] = KEY_GAINDECR;
							cmd[1] = volstep;		// VOLUME_STEP;
		    				writeall(fdplayer[0], cmd, 2);
		    				Volume -= volstep;		// VOLUME_STEP;
		    				if (send_now_playing == 0) send_now_playing = 1;
		    			}
					} else if (PinEvent == WINCTRL_CMD_NEXT ){
						cmd[0] = KEY_QUIT;
	    				writeall(fdplayer[0], cmd, 1);
	    				ret = PLAY_NEXT_MP3;
					} else if (PinEvent == WINCTRL_CMD_PREV ){
						cmd[0] = KEY_QUIT;
	    				writeall(fdplayer[0], cmd, 1);
	   					ret = PLAY_PREV_MP3;
					} else if (PinEvent == WINCTRL_CMD_PLAYSKIP ){
						cmd[0] = KEY_QUIT;
	    				writeall(fdplayer[0], cmd, 1);
	    				ret = PLAY_OTHER_DIR;
					} else if (PinEvent == WINCTRL_CMD_PLAYSTOP ){
						cmd[0] = KEY_QUIT;
	    				writeall(fdplayer[0], cmd, 1);
	    				ret = SELECT_OTHER_MP3;
					} else if (PinEvent == WINCTRL_CMD_SHUTDOWN){
						cmd[0] = KEY_QUIT;
	    				writeall(fdplayer[0], cmd, 1);
	    				ret = POWEROFF;
						break;
					} else if (PinEvent == WINCTRL_CMD_FILESIZE){
						SendCmdToCtrlFilesize((unsigned int)mp3stat.st_size);
					} else if (PinEvent == WINCTRL_CMD_ACTIVE){
	    				if (send_now_playing == 0) send_now_playing = 1;
					} else if (PinEvent == WINCTRL_CMD_DATATIMESTAMP){
						SendDataTimestamp();
					} else {
						PinEvent = 0;
					}
				}
				//if (PinEvent != 0){
				//	printf("PinEvent=%d\n", PinEvent);
				//}
				PinEvent = 0;
			}
			
#if 0		// 未完 (2019/12/08)
			// PICからの終了指示
			if (digitalRead(PIN_INP_PDOWN) == 0){
				putchar('0');
				cmd[0] = KEY_QUIT;
				writeall(fdplayer[0], cmd, 1);
				ret = POWEROFF;
				break;
			}
#endif
			
			current_systime_ms = millis();
			if (current_systime_ms > AutoPoffTimer){
				cmd[0] = KEY_QUIT;
				writeall(fdplayer[0], cmd, 1);
				ret = POWEROFF;
				break;		// 明示的にbreakしないと無限ループになる？
			} else if (send_now_playing == 1){
				// madplayに現在再生中のファイル位置を要求
				cmd[0] = KEY_GETPOS;
				writeall(fdplayer[0], cmd, 1);
				send_now_playing = 2;
			}
			
			// madplayからの受信
			if (send_now_playing == 2){
				int	rval = 0;
				unsigned char buf[CMD_BUF_SIZE];
				struct pollfd fd_poll[1];
				int	ok_flg = 0;
				int	retry_cnt = 0;
				int total = 0;
				int	wsz;
				
				do {
					memset(&fd_poll, 0, sizeof(fd_poll));
					fd_poll[0].fd = fdplayer[0];
					fd_poll[0].events = POLLIN | POLLERR;
					poll(fd_poll, 1, POLL_TIMEOUT);
					if (fd_poll[0].revents & POLLIN) {
						//rval = readall(fdplayer[0], buf, 5);	// 5
						wsz = read(fdplayer[0], buf + total, PLAYER_CMDLENGTH - total);
						if (wsz > 0){
							total += wsz;
							if (wsz < PLAYER_CMDLENGTH){
								printf("partial read %d \n", wsz);
							}
						//if (rval > 0){
							if (buf[0] == KEY_GETPOS && total == PLAYER_CMDLENGTH){
								//unsigned int total_play_time = 0;
								//unsigned int current_play_time = 0;
								current_play_time = (unsigned int)buf[1];
								current_play_time |= ((unsigned int)buf[2] << 8);
								
								total_play_time = (unsigned int)buf[3];
								total_play_time |= ((unsigned int)buf[4] << 8);
								// ウインドウズ版コントローラへの送信
								SendCmdToCtrlNowPlaying(dirno, trackno, current_play_time, Volume, NOWPLAYING_CONTINUE, total_play_time);
								printf("play=%d total=%d\n", current_play_time, total_play_time);
								// 
								if ((total_play_time - current_play_time) <= 10){
									cmd_acceptable = 0;		// 曲の終わり際はコントローラからのコマンドをデコーダに送らない
								}
								ok_flg = 1;
							}
						} else {
							break;
						}
					}
					retry_cnt++;
				} while (ok_flg == 0 || retry_cnt < 2);
				if (ok_flg == 0){
					printf("NO position\n");
				}
				send_now_playing = 0;
			}
			
			// madplayの実行状態
			if ((pid = waitpid(pid, &status, WNOHANG)) == -1){
				printf("waitpid error");
				break;
			} else if (pid != 0) {
				//sleep(1);
			//}
			//else {
				if (WIFEXITED(status))
					printf("child exited with status of %d\r¥n", WEXITSTATUS(status));
				else puts("child did not exit successfully");
			}
		} while (pid == 0);
		close(fdplayer[0]);
		close(fdplayer[1]);
		
		printf("madplay end\n");
		
		break;
	}
	
	if (ret == PLAY_ONCEMORE){
		goto restart_label;
	} else {
		start_time += (millis() - start_systime_ms)/1000;		// 経過時間(秒)
		memcpy(PlayDirHist + HS_LASTTIME_OFFSET, &start_time, HS_LASTTIME_LEN);
		SaveHistory();
	}
	
	return ret;
}


/*
 *	MP3ファイルが存在しないことをIR01に通知
 */
void SendCmdNoMp3_IR01(char exist_flg){
	char	buf[WINCTRL_CMDLENGTH];
	
	memset(buf, 0, WINCTRL_CMDLENGTH);
	buf[0] = IR01_CMD_EXISTMP3;
	buf[1] = exist_flg;
	writeall(fdWin[0], buf, WINCTRL_CMDLENGTH);
}

/*
 *	IR01専用ディレクトリ内のMP3ファイルをIR01からの指示で再生
 */
int PlayOneMp3_IR01(const char *mp3file, uint32_t start_time){
	pid_t	pid;
	int 	status;
	int		fdplayer[2];
	char	cmd[CMD_BUF_SIZE];
	char	fdstr[24];
	char	attenuate[24];
	char	start[24];
	//char	quiet[16];
	char	notty[24];
	//char	displaytime[32];
	char	*args[16];			// madplay起動時引数配列
	int		argidx;
    int		ret;
    int		volstep;
	unsigned int	start_systime_ms;		// 再生開始システム時刻(ミリ秒：millis()の戻り値)
	unsigned int	current_systime_ms;		// 現在システム時刻(ミリ秒：millis()の戻り値)
	unsigned int 	total_play_time;		// 曲の長さ（秒）
	unsigned int 	current_play_time;		// 現在の再生秒
	int		send_now_playing;
	struct stat		mp3stat;				// 
	int		cmd_acceptable;					// 1:デコーダへのコマンド可能
	int		i;
    
	if (stat(mp3file, &mp3stat) < 0){
		SendCmdNoMp3_IR01(2);
		return 0;
	} else {
		SendCmdNoMp3_IR01(1);
	}
	
	if (socketpair(AF_LOCAL, SOCK_STREAM, 0, fdplayer) == -1){
		printf("socketpair error %d\n", errno);
		return -1;
	}
	
	pid = fork();
	switch(pid){
	case -1:
		exit(1);
		break;
	case 0:
		argidx = 0;
		args[argidx++] = "madplay";
		args[argidx++] = mp3file;
		sprintf(fdstr, "--socket=%d", fdplayer[1]);
		args[argidx++] = fdstr;
		if (Volume < 0){
			if (Volume < MIN_VOLUME) Volume = MIN_VOLUME;
			sprintf(attenuate, "--attenuate=%d", Volume);
			args[argidx++] = attenuate;
		} else if (Volume > 0){
			if (Volume > MAX_VOLUME) Volume = MAX_VOLUME;
			sprintf(attenuate, "--amplify=%d", Volume);
			args[argidx++] = attenuate;
		}
		{
			int h, m, s;
			h = (start_time / 3600);			// hour
			m = (start_time - h*3600) / 60;	// min
			s = (start_time % 60);
			sprintf(start, "--start=%02d:%02d:%02d.000", h, m, s);
			args[argidx++] = start;
		}
		sprintf(notty, "--no-tty-control");
		args[argidx++] = notty;
		args[argidx++] = (char *)0;
		execv("/usr/local/bin/madplay", args);
		_exit(errno);
		break;
	default:
//TODO:		SendCmdToCtrlNowPlaying(dirno, trackno, start_time, Volume, NOWPLAYING_STARTEDNOW, 0);
		digitalWrite(PIN_OUT_MUTE, MUTE_OUTPUT_ON);
		ret = 0;
		PinEvent = 0;
		send_now_playing = 0;
		cmd_acceptable = 1;		// デコーダにコマンド送信可能
		
		do {
			unsigned char	buf[CMD_BUF_SIZE];
			
			// ウインドウズ版コントローラからの受信
			if (ReadWindowsController(buf, WINCTRL_CMDLENGTH) > 0 && ret == 0){
				// retの値はMP3_14/decoder.c(run_sync())に合わせる
				if (cmd_acceptable == 1){
					PinEvent = buf[0];
					//printf("cmd from client = %x\n", PinEvent);
					if (PinEvent == WINCTRL_CMD_VOLUP ){
						//volstep = MAX_VOLUME - Volume;
						//if (volstep > VOLUME_STEP) volstep = VOLUME_STEP;
						volstep = buf[1];
						if ((Volume + volstep) > MAX_VOLUME) volstep = MAX_VOLUME - Volume;
						if (volstep > 0){
							cmd[0] = KEY_GAININCR;
							cmd[1] = volstep;			// VOLUME_STEP
		    				writeall(fdplayer[0], cmd, 2);
		    				Volume += volstep;			// VOLUME_STEP;
		    				if (send_now_playing == 0) send_now_playing = 1;
		    			}
					} else if (PinEvent == WINCTRL_CMD_VOLDOWN ){
						//volstep = Volume - MIN_VOLUME;
						//if (volstep > VOLUME_STEP) volstep = VOLUME_STEP;
						volstep = buf[1];
						if ((Volume - volstep) < MIN_VOLUME) volstep = Volume - MIN_VOLUME;
						if (volstep > 0){
							cmd[0] = KEY_GAINDECR;
							cmd[1] = volstep;		// VOLUME_STEP;
		    				writeall(fdplayer[0], cmd, 2);
		    				Volume -= volstep;		// VOLUME_STEP;
		    				if (send_now_playing == 0) send_now_playing = 1;
		    			}
					} else if (PinEvent == WINCTRL_CMD_PLAYSTOP ){
						cmd[0] = KEY_QUIT;
	    				writeall(fdplayer[0], cmd, 1);
	    				ret = SELECT_OTHER_MP3;
					} else if (PinEvent == WINCTRL_CMD_SHUTDOWN){
						cmd[0] = KEY_QUIT;
	    				writeall(fdplayer[0], cmd, 1);
	    				ret = POWEROFF;
						break;
					} else if (PinEvent == WINCTRL_CMD_ACTIVE){
	    				if (send_now_playing == 0) send_now_playing = 1;
					} else {
						PinEvent = 0;
					}
				}
				//if (PinEvent != 0){
				//	printf("PinEvent=%d\n", PinEvent);
				//}
				PinEvent = 0;
			}
			
			if (send_now_playing == 1){
				// madplayに現在再生中のファイル位置を要求
				cmd[0] = KEY_GETPOS;
				writeall(fdplayer[0], cmd, 1);
				send_now_playing = 2;
			}
			
			// madplayからの受信
			if (send_now_playing == 2){
				int	rval = 0;
				unsigned char buf[CMD_BUF_SIZE];
				struct pollfd fd_poll[1];
				int	ok_flg = 0;
				int	retry_cnt = 0;
				int total = 0;
				int	wsz;
				
				do {
					memset(&fd_poll, 0, sizeof(fd_poll));
					fd_poll[0].fd = fdplayer[0];
					fd_poll[0].events = POLLIN | POLLERR;
					poll(fd_poll, 1, POLL_TIMEOUT);
					if (fd_poll[0].revents & POLLIN) {
						//rval = readall(fdplayer[0], buf, 5);	// 5
						wsz = read(fdplayer[0], buf + total, PLAYER_CMDLENGTH - total);
						if (wsz > 0){
							total += wsz;
							if (wsz < PLAYER_CMDLENGTH){
								printf("partial read %d \n", wsz);
							}
						//if (rval > 0){
							if (buf[0] == KEY_GETPOS && total == PLAYER_CMDLENGTH){
								//unsigned int total_play_time = 0;
								//unsigned int current_play_time = 0;
								current_play_time = (unsigned int)buf[1];
								current_play_time |= ((unsigned int)buf[2] << 8);
								
								total_play_time = (unsigned int)buf[3];
								total_play_time |= ((unsigned int)buf[4] << 8);
								// コントローラへの送信
								SendCmdToCtrlNowPlaying(0, 0, current_play_time, Volume, NOWPLAYING_CONTINUE, total_play_time);
								printf("play=%d total=%d\n", current_play_time, total_play_time);
								// 
								if ((total_play_time - current_play_time) <= 10){
									cmd_acceptable = 0;		// 曲の終わり際はコントローラからのコマンドをデコーダに送らない
								}
								ok_flg = 1;
							}
						} else {
							break;
						}
					}
					retry_cnt++;
				} while (ok_flg == 0 || retry_cnt < 2);
				if (ok_flg == 0){
					printf("NO position\n");
				}
				send_now_playing = 0;
			}
			
			// madplayの実行状態
			if ((pid = waitpid(pid, &status, WNOHANG)) == -1){
				printf("waitpid error");
				break;
			} else if (pid != 0) {
				//sleep(1);
			//}
			//else {
				if (WIFEXITED(status))
					printf("child exited with status of %d\r¥n", WEXITSTATUS(status));
				else puts("child did not exit successfully");
			}
		} while (pid == 0);
		close(fdplayer[0]);
		close(fdplayer[1]);
		digitalWrite(PIN_OUT_MUTE, MUTE_OUTPUT_OFF);
		
		printf("madplay end\n");
		
		break;
	}
	
	return ret;
}


/*
 *	ウインドウズ版コントローラへWINCTRL_CMD_NOWPLAYINGの送信
 */
void SendCmdToCtrlNowPlaying(uint16_t dirno, uint8_t trackno, unsigned int current_play_time, int volume, uint8_t flag, unsigned int position){
	char	buf[WINCTRL_CMDLENGTH];
	
//	printf("SendCmdToCtrlNowPlaying dirno=%d trackno=%d\n", dirno, trackno);
	buf[0] = WINCTRL_CMD_NOWPLAYING;
	buf[1] = (dirno & 0xFF);
	buf[2] = ((dirno >> 8) & 0xFF);
	buf[3] = trackno;
	buf[4] = (current_play_time & 0xFF);
	buf[5] = ((current_play_time >> 8) & 0xFF);
	buf[6] = ((unsigned int)volume & 0xFF);
	buf[7] = (((unsigned int)volume >> 8) & 0xFF);
	buf[8] = (((unsigned int)volume >> 16) & 0xFF);
	buf[9] = (((unsigned int)volume >> 24) & 0xFF);
	buf[10] = flag;
	buf[11] = (position & 0xFF);
	buf[12] = ((position >> 8) & 0xFF);
	buf[13] = 0;
	buf[14] = 0;
	writeall(fdWin[0], buf, WINCTRL_CMDLENGTH);
}


/*
 * ウインドウズ版コントローラへWINCTRL_CMD_FILESIZEの送信
 */
void SendCmdToCtrlFilesize(unsigned int filesize){
	char	buf[WINCTRL_CMDLENGTH];
	buf[0] = WINCTRL_CMD_FILESIZE;
	buf[1] = (filesize & 0xFF);
	buf[2] = ((filesize >> 8) & 0xFF);
	buf[3] = ((filesize >> 16) & 0xFF);
	buf[4] = ((filesize >> 24) & 0xFF);
	writeall(fdWin[0], buf, WINCTRL_CMDLENGTH);
}


/*
 *	所定バイトをすべてwriteするまで繰り返す
 */
int writeall(int fd, void *buf, size_t nbyte){
	ssize_t	wsz = 0;
	int		total = 0;
	
	do {
		wsz = write(fd, buf + total, nbyte - total);
		if (wsz >= 0){
			total += wsz;
		} else {
			return total;
		}
	} while (total < nbyte);
	
	return total;
}


/*
 *	所定バイトをすべてreadするまで繰り返す
 */
int readall(int fd, void *buf, size_t nbyte){
	ssize_t	wsz = 0;
	int		total = 0;
	
	do {
		wsz = read(fd, buf + total, nbyte - total);
		if (wsz >= 0){
			total += wsz;
		} else {
			return total;
		}
	} while (total < nbyte);
	
	return total;
}


/*
 * 履歴ファイルを読み込む
 */
void ReadHistory(){
	FILE		*Fil;

	Fil = fopen(HISTFILE, "r");
	if (Fil != NULL){
		fread(&PlayDirHist, RECLEN_HISTFILE, 1, Fil);
		fclose(Fil);
	}
}

/*
 * 履歴ファイルを保存
 */
void SaveHistory(){
	FILE		*Fil;
	
	Fil = fopen(HISTFILE, "w");
	if (Fil != NULL){
		fwrite(&PlayDirHist, RECLEN_HISTFILE, 1, Fil);
		fclose(Fil);
		
		printf("SaveHistory()\n");
	}
	
}


/*
 *	mp3_14_mkflを実行
 */
void ExecMkfl(unsigned char *param){
	char	arg[16];
	char	cmd[128];
	
	sprintf(cmd, FIND_SND_ROOT, SND_ROOT, DIRLISTFILE);
	system(cmd);
	
	memcpy(arg, param, 14);		// タイムスタンプが渡される
	arg[14] = '\0';
	sprintf(cmd, MAKE_SNDFILE_LIST, arg);
	system(cmd);
	
	// 終了したことをコントローラに通知
	memset(cmd, 0, WINCTRL_CMDLENGTH);
	cmd[0] = WINCTRL_CMD_EXECMKFL;
	writeall(fdWin[0], cmd, WINCTRL_CMDLENGTH);
}


/*
 *	HOST上のデータファイルのタイムスタンプを取得
 */
void ReadDataTimestamp(uint8_t *RecDirList){
	FILE	*Fil;
	int		i;

	memset(DataTimeStamp, '0', DL_CRDATETIME_LEN);
	
	Fil = fopen(DIRSTRUCTUREFILE14, "r");
	if (Fil != NULL){
		if (fread(RecDirList, 1, RECLEN_DIRLIST_FULL, Fil) == RECLEN_DIRLIST_FULL){
			for (i=0; i<DL_CRDATETIME_LEN; i++){
				DataTimeStamp[i] = RecDirList[4 + i];
			}
		}
	    fclose(Fil);
	}
}


/*
 *	HOST上のデータファイルのタイムスタンプをコントローラに返す
 */
void SendDataTimestamp(){
	char	cmd[WINCTRL_CMDLENGTH];
	int		i;
	
	printf("SendDataTimestamp start\n");
	memset(cmd, '0', WINCTRL_CMDLENGTH);
	cmd[0] = WINCTRL_CMD_DATATIMESTAMP;
	for (i=0; i<DL_CRDATETIME_LEN; i++){
		cmd[1 + i] = DataTimeStamp[i];
	}
	writeall(fdWin[0], cmd, WINCTRL_CMDLENGTH);
	printf("SendDataTimestamp end \n");
}


